﻿public abstract class ReservableEntity {
	public Driver? ReservedByDriver { get; private set; } = null;

	public bool ReserveFor(Driver driver) {
		if (ReservedByDriver is not null) {
			return false;
		}
		ReservedByDriver = driver;
		return true;
	}

	public void CancelReservationFor(Driver driver) {
		if (ReservedByDriver == driver) {
			ReservedByDriver = null;
		}
	}
}

public enum PointPosition { None, Normal, Reverse }

public static class PointPositionExtensions {
	public static char ToSymbolChar(this PointPosition position) {
		return position switch {
			PointPosition.None => '?',
			PointPosition.Normal => '+',
			PointPosition.Reverse => '-',
			_ => '~'
		};
	}
}

public class PointDrive {
	public string Name { get; set; } = string.Empty;
	public Point Point { get; set; } = null!;
}

public class Point {
	public required string Name { get; set; }
	public PointPosition Position { get; set; } = PointPosition.Normal;
	public PointDrive[] Drives { get; set; } = Array.Empty<PointDrive>();

	private int _routeLocks = 0;

	public void AddRouteLock() { _routeLocks++; }

	public void RemoveRouteLock() { _routeLocks--; }

	public bool IsRouteLocked => _routeLocks > 0;
}

public class PointGroup : ReservableEntity {
	public required string Name { get; set; }
	public required Point[] Points { get; set; }

	public bool IsInConsistentPosition(PointPosition targetPosition) => Points.All(p => p.Position == targetPosition);
}

public record struct RequiredPointPosition {
	public PointGroup PointGroup { get; set; }
	public PointPosition Position { get; set; }
}

public class TvpSection : ReservableEntity {
	public required string Name { get; set; }
	public bool IsVacant { get; set; } = true;
	public bool IsRouteLocked { get; private set; } = false;
	public PointGroup[] PointGroups = Array.Empty<PointGroup>();

	public void SetRouteLock() {
		IsRouteLocked = true;
		Array.ForEach(PointGroups, pg =>
			Array.ForEach(pg.Points, p => p.AddRouteLock())
		);
	}

	public void ReleaseRouteLock() {
		Array.ForEach(PointGroups, pg =>
			Array.ForEach(pg.Points, p => p.RemoveRouteLock())
		);
		IsRouteLocked = false;
	}
}

public class Signal {
	public required string Name { get; init; }
}

public enum RouteType { Main, Shunting }

public class Route {
	public int Id { get; set; }
	public RouteType Type { get; set; }
	public required Signal EntrySignal { get; set; }
	public RequiredPointPosition[] RequiredPointPositions { get; set; } = Array.Empty<RequiredPointPosition>();
	public RequiredPointPosition[] FlankProtectionPointPositions { get; set; } = Array.Empty<RequiredPointPosition>();
	public TvpSection[] SectionsInRoute { get; set; } = Array.Empty<TvpSection>();
	public TvpSection[] SectionsRequiredVacant { get; set; } = Array.Empty<TvpSection>();
}

public class InterlockingTable {
	public required Route[] Routes { get; set; }
}