﻿public abstract class CoroutineStepResult {
	public static CoroutineStepSuccessful Success() => new CoroutineStepSuccessful();
	public static CoroutineStepError Error(string message) => new CoroutineStepError { ErrorMessage = message };
	public static CoroutineStepError Error<TElement>(string message, TElement associatedElement) => new CoroutineStepError<TElement> { ErrorMessage = message, AssociatedElement = associatedElement };
}

public sealed class CoroutineStepSuccessful : CoroutineStepResult { }

public class CoroutineStepError : CoroutineStepResult {
	public required string ErrorMessage { get; init; }
}

public sealed class CoroutineStepError<TElement> : CoroutineStepError {
	public required TElement AssociatedElement { get; init; }
}

public enum CoroutinePriority { PreDriver = 0, Driver, PostDriver, Last };

public class CoroutineScheduler {
	public static CoroutineScheduler Current { get; set; } = new CoroutineScheduler();

	public int CurrentTime { get; private set; }
	public FieldObjectMessenger? Messenger { get; set; }

	public List<IEnumerator<CoroutineStepResult>>[] CoroutineQueues { get; } = new List<IEnumerator<CoroutineStepResult>>[(int) CoroutinePriority.Last + 1];
	public Driver?[] Drivers { get; } = new Driver[5];
	public List<string> Errors { get; } = new List<string>() { "", "", "", "", "", "", "" };

	public CoroutineScheduler() {
		for (int i = 0; i < CoroutineQueues.Length; i++) {
			CoroutineQueues[i] = new List<IEnumerator<CoroutineStepResult>>();
		}
	}

	public bool Enqueue(Driver driver) {
		for (int i = 0; i < Drivers.Length; i++) {
			if (Drivers[i] is null) {
				Drivers[i] = driver;
				driver.StartTime = CurrentTime;
				return true;
			}
		}
		return false;
	}

	public void Enqueue(IEnumerator<CoroutineStepResult> coroutine, CoroutinePriority priority) {
		CoroutineQueues[(int) priority].Add(coroutine);
	}

	public void Run() {
		var startTime = DateTime.Now;

		while (true) {
			CurrentTime = (int) (DateTime.Now - startTime).TotalMilliseconds;

			for (int queueId = 0; queueId < CoroutineQueues.Length; queueId++) {
				if (queueId == (int) CoroutinePriority.Driver) {
					// Driver priority coroutine queue is not used - there is fixed number of driver slots, and drivers have a main coroutine (executed in slot order) + OnRemovingFromScheduler callback:

					for (int i = 0; i < Drivers.Length; i++) {
						if (Drivers[i] is not null) {
							Driver driver = Drivers[i]!;
							bool endDriver = false;
							if (!driver.MainCoroutine.MoveNext()) {
								endDriver = true;
							} else if (driver.MainCoroutine.Current is CoroutineStepError error) {
								Errors.RemoveAt(0);
								Errors.Add(error.ErrorMessage);
								endDriver = true;
							}

							if (endDriver) {
								driver.OnRemovingFromScheduler();
								Drivers[i] = null;
							}
						}
					}
				} else {
					// All non-Driver priorities have plain queues - execute all coroutines from current priority level queue in order of coroutines registration:

					var queue = CoroutineQueues[queueId];
					for (int i = 0; i < queue.Count; i++) {
						var coroutine = queue[i];
						bool removeCoroutine = false;
						if (!coroutine.MoveNext()) {
							removeCoroutine = true;
						} else if (coroutine.Current is CoroutineStepError error) {
							removeCoroutine = true;
						}

						if (removeCoroutine) {
							queue.RemoveAt(i);
							i--;
						}
					}
				}
			}

			Thread.Sleep(100);
		}
	}
}

public struct TimeoutGuard {
	private int _endTime;

	public TimeoutGuard(int timeout) {
		_endTime = CoroutineScheduler.Current.CurrentTime + timeout;
	}

	public bool HasExpired => CoroutineScheduler.Current.CurrentTime >= _endTime;
}

public abstract class Driver {
	public IEnumerator<CoroutineStepResult> MainCoroutine { get; }

	public int StartTime { get; set; }

	public Driver() {
		// NOTE: GetMainCoroutine() should be iterator method, so that it captures only "this", and is not using descendant fields/properties initialized later.
		MainCoroutine = GetMainCoroutine();
	}

	protected abstract IEnumerator<CoroutineStepResult> GetMainCoroutine();

	public virtual void OnRemovingFromScheduler() { }
}