﻿public class CallRouteDriver : Driver {
	private Route _route;

	public CallRouteDriver(Route route) {
		_route = route;
	}

	public Route Route => _route;

	protected override IEnumerator<CoroutineStepResult> GetMainCoroutine() {
		// DEBUG: Yield return immediatelly so that we see the driver in slot for the first cycle in Dumper output - this is not by real implementation.
		yield return CoroutineStepResult.Success();

		// Reserve all route body entities:
		bool allReserved = _route.SectionsInRoute.All(s => s.ReserveFor(this));
		if (allReserved) {
			allReserved = _route.RequiredPointPositions.All(rpp => rpp.PointGroup.ReserveFor(this));	// FIXME: Only the actual part of the PointGroup in the route body should be reserved here, but we have not enough topological data in the model in this version for simplicity.
		}
		if (allReserved) {
			allReserved = _route.FlankProtectionPointPositions.All(rpp => rpp.PointGroup.ReserveFor(this));
		}
		if (!allReserved) {
			yield return CoroutineStepResult.Error("Blok. prováděnou volbou");
		}

		// Give user initial time to cancel the request
		var timeout = new TimeoutGuard(500);
		while (!timeout.HasExpired) {
			yield return CoroutineStepResult.Success();
		}

		// Check route body TVP sections not route locked:
		foreach (var section in _route.SectionsInRoute) {
			if (section.IsRouteLocked) {
				yield return CoroutineStepResult.Error("Kolejový úsek zapevněn", section);
			}
		}

		// Check route body TVP sections are vacant:
		foreach (var section in _route.SectionsInRoute) {
			if (!section.IsVacant) {
				yield return CoroutineStepResult.Error("Obsazený kolejový úsek", section);
			}
		}

		// Check points not in required position are not route locked:
		foreach (var requiredPointPosition in _route.RequiredPointPositions.Concat(_route.FlankProtectionPointPositions)) {
			foreach (var point in requiredPointPosition.PointGroup.Points.Where(p => p.Position != requiredPointPosition.Position)) {
				if (point.IsRouteLocked) {
					yield return CoroutineStepResult.Error("Závěr výhybky", point);
				}
			}
		}

		// Throw points:
		ThrowPoints(_route.RequiredPointPositions);
		while (!CheckAllPointsInPosition(_route.RequiredPointPositions)) {
			yield return CoroutineStepResult.Success();
		}

		// Throw flank points:
		ThrowPoints(_route.FlankProtectionPointPositions);
		while (!CheckAllPointsInPosition(_route.FlankProtectionPointPositions)) {
			yield return CoroutineStepResult.Success();
		}

		#region TODO: implement rest of the call route algorithm
		// TODO: Request lock TVP sections in route (executed by lower levels) -> wait for all sections locks & possibly timeout with "Nezapevněná jízdní cesta" 
		// (locked sections will be indicated as "cyan")

		// TODO: Lock all flank points -> wait for all point locks

		// TODO: Mark TVP sections required vacant as used by a route (indicated as "yellow") - this is not a section lock

		// TODO: Check route consistency (all sections and points locked) -> mark all sections as main route and assign route to entry signal
		// (locked sections part of main route will be indicated as "lime green" now)

		// TODO: Request for clearing entry signal ("green" light) -> wait for correct aspect indicated & possibly timeout with "Nerozsvícení pov.návěsti"
		#endregion
	}

	private static void ThrowPoints(RequiredPointPosition[] requiredPointPositions) {
		foreach (var requiredPointPosition in requiredPointPositions) {
			var messenger = (MockFieldObjectMessenger) CoroutineScheduler.Current.Messenger!;

			foreach (var point in requiredPointPosition.PointGroup.Points) {
				if (point.Position != requiredPointPosition.Position) {
					messenger.ThrowPoint(point, requiredPointPosition.Position);
				}
			}
		}
	}

	private static bool CheckAllPointsInPosition(RequiredPointPosition[] requiredPointPositions) {
		foreach (var requiredPointPosition in requiredPointPositions) {
			foreach (var point in requiredPointPosition.PointGroup.Points) {
				if (point.Position != requiredPointPosition.Position) {
					return false;
				}
			}
		}
		return true;
	}

	public override void OnRemovingFromScheduler() {
		// Cancel all reservations:
		Array.ForEach(_route.FlankProtectionPointPositions, rpp => rpp.PointGroup.CancelReservationFor(this));
		Array.ForEach(_route.RequiredPointPositions, rpp => rpp.PointGroup.CancelReservationFor(this));
		Array.ForEach(_route.SectionsInRoute, s => s.CancelReservationFor(this));
	}
}

public class CancelRouteDriver : Driver {
	private Route _route;

	public CancelRouteDriver(Route route) {
		_route = route;
	}

	public Route Route => _route;

	protected override IEnumerator<CoroutineStepResult> GetMainCoroutine() {
		// Release point locks:
		Array.ForEach(_route.FlankProtectionPointPositions, rpp =>
			Array.ForEach(rpp.PointGroup.Points, p => p.RemoveRouteLock())
		);

		yield return CoroutineStepResult.Success();

		// TODO: Implement all the steps for route release (release TVP sections, etc.)
	}
}

public class SetPointGroupPositionDriver : Driver {
	private PointGroup _pointGroup;
	private PointPosition _targetPosition;

	public SetPointGroupPositionDriver(PointGroup pointGroup, PointPosition targetPosition) {
		_pointGroup = pointGroup;
		_targetPosition = targetPosition;
	}

	public PointGroup PointGroup => _pointGroup;
	public PointPosition TargetPosition => _targetPosition;

	// TODO: Add support for changing targetPosition by user command during driver being active (i.e. during scheduled main coroutine)

	protected override IEnumerator<CoroutineStepResult> GetMainCoroutine() {
		// DEBUG: Yield return immediatelly so that we see the driver in slot for the first cycle in Dumper output - this is not by real implementation.
		yield return CoroutineStepResult.Success();

		if (!_pointGroup.ReserveFor(this)) {
			yield return CoroutineStepResult.Error("Blok. prováděnou volbou");
		}

		// TODO: Check points are not route locked!

		var messenger = (MockFieldObjectMessenger) CoroutineScheduler.Current.Messenger!;
		foreach (var point in _pointGroup.Points) {
			if (point.Position != _targetPosition) {
				messenger.ThrowPoint(point, _targetPosition);
			}
		}

		while (!CheckAllPointsInPosition()) {
			yield return CoroutineStepResult.Success();
		}
	}

	private bool CheckAllPointsInPosition() {
		foreach (var point in _pointGroup.Points) {
			if (point.Position != _targetPosition) {
				return false;
			}
		}
		return true;
	}

	public override void OnRemovingFromScheduler() {
		_pointGroup.CancelReservationFor(this);
	}
}
